-- -*-coding: utf-8;-*-

--[[
-- ************************************************************************
-- BuddyName.lua
-- ************************************************************************
-- This addon is released under GPL v3. All additionally used libraries are
-- released under the license that accompanies them.
--]]

BuddyName = LibStub("AceAddon-3.0"):NewAddon("BuddyName",
                                             "AceConsole-3.0",
                                             "AceHook-3.0",
                                             "AceTimer-3.0",
                                             "AceComm-3.0");

local L = LibStub("AceLocale-3.0"):GetLocale("BuddyName");

local Compressor = LibStub:GetLibrary("LibCompress");

local chatChannels = {
	 ["WHISPER"]      = true,
	 ["PARTY"]        = true,
	 ["GUILD"]        = true,
	 ["YELL"]         = true,
	 ["PARTY"]        = true,
	 ["RAID"]         = true,
	 ["BATTLEGROUND"] = true,
	 ["CHANNEL"]      = false,
}

local slashCommands = {
	 ["bn"]        = true,
	 ["buddyname"] = true,
	 ["BuddyName"] = true,
}

-- perhaps used in future versions
local registerEvents = {
	 ["GUILD_JOIN"] = false,
}

-- Constants used for synchronization
local COMM = {
	 SYNC_PREFIX = "BuddyName",
	 PROTOCOL = 2,
	 SEP = { -- separator
			SECTION = "|",
			ENTRY   = ";",
			FIELD   = ":",
	 },
	 SECTION_TYPE = {
			ENTRY_SET        = 1,
			HASH_PROPAGATION = 2,
			DATA_REQUEST     = 3,
			UNKNOWN          = -1,
	 },
	 HEADER = {
			PREFIX="BN",
			COMPRESSION = {
				 NONE        = 0,
				 LIBCOMPRESS = 1,
			},
	 },
	 PATTERNS = {}, -- filled in below
};
COMM.PATTERNS = {
	 SEARCH = {
			HEADER   = "^"..COMM.HEADER.PREFIX..COMM.SEP.ENTRY.."(%d+)"..COMM.SEP.ENTRY.."(%d+)"..COMM.SEP.SECTION,
			SECTION_TYPE = "^(%d+)"..COMM.SEP.ENTRY,
	 },
	 BUILD = {
			HEADER = COMM.HEADER.PREFIX..COMM.SEP.ENTRY..COMM.PROTOCOL..COMM.SEP.ENTRY..COMM.HEADER.COMPRESSION.LIBCOMPRESS..COMM.SEP.SECTION,
			ENTRY_SET_STUB = COMM.SECTION_TYPE.ENTRY_SET,
	 },
};

--[[ Local variables ]]
local BN_DB_Version = 2; -- Database format
local BN_Active = false; -- Is this addon active? Modified by OnEnable and OnDisable
local BN_EventFrame = nil; -- Used to register events and process them. Created in OnInitialize
local currentRealm = nil;
local BN_Timer = {}; -- table for storing active timers
local warnedAboutUpdate = false; -- whether the user has been warned about an existing update this session
local compressType = 1; -- 0 means no compression, 1 means usage of LibCompression
local commActive = false; -- debug, remove me later
local debugMode = false; -- whether to output verbose debugging messages

--[[ -- -------------------------------- -- ]]
--[[ -- Basic functions, no dependenices -- ]]
--[[ -- -------------------------------- -- ]]

--[[ Creates a new numeric timestamp for the current time and date
--   Returns:
--     1. Numeric: timestamp in format YYYYMMDDHHmm
--]]
function BuddyName:GetTimestamp()
	 local _,month,day,year = CalendarGetDate();
	 local hour,minute = GetGameTime();

	 local timestamp = minute + 100 * (hour + 100 * (day + 100 * (month + 100 * year)))

	 return timestamp
end

--[[ Concatenates an arbitrary number of arguments to one string.
--   Only used for debugging purposes.
--   Returns:
--     1. Concatenated version of all supplied arguments
--]]
local function allToString(...)
	 return strjoin(" ", tostringall(...));
end

local function PlayerIsInGuild()
	 local guild = GetGuildInfo("player");
	 if (guild) then
			return true;
	 else
			return false;
	 end
end

local function debug(...)
	 if (debugMode==true) then
			local msg="|cFF55FFFFBuddyName(|cFFFF0000Debug|r):|r "..strjoin(":", tostringall(...));
			print(msg);
	 end
end

--[[ -- -------------------------- -- ]]
--[[ -- Database-related functions -- ]]
--[[ -- -------------------------- -- ]]

--[[ Checks a given name on whether it is valid or not.
--   This means especially that checks are performed whether the
--   name contains one of the separation chars from synchronization.
--   However, it does NOT check whether a player with this name is
--   online or even exists.
--   Returns:
--     1. true if name is valid; false else
--]]
local function IsValidName(name)
	 local checkString = "["..COMM.SEP.SECTION..COMM.SEP.ENTRY..COMM.SEP.FIELD.."]";

	 local pstart = string.find(name,checkString);
	 if (pstart) then
			return false;
	 else
			return true;
	 end

	 -- TODO: distinguish between char name and buddy name.
	 -- Buddy names may contain spaces or numbers, char names may not.
	 -- Also, auto-format charnames: first letter uppercase, other letters not.
end

--[[ Checks whether the given charname has a buddy name in our DB
--   Returns:
--     1. Boolean: true if author was found in DB, false else
--     2. String: Buddy name of charname, or nil if it doesn't exist or equals empty string
--]]
function BuddyName:IsInDB(charname)
	 if (charname~=nil and charname~="") then
			if (self.db.realm.char and -- simplify? test!
					self.db.realm.char[charname] and
					self.db.realm.char[charname].buddy and
					self.db.realm.char[charname].buddy~="")
			then
				 return true,self.db.realm.char[charname].buddy;
			end
	 end

	 return false, nil
end

--[[ Adds a new entry to the DB or updates an existing one.
--   If a timestamp is provided, then this will be used.
--   Returns:
--     1. Boolean: true if successul, false else
--]]
function BuddyName:addToDB(charname, buddyname, timestamp)
	 if (charname==nil or buddyname==nil) then
			-- output error message
			return false;
	 end

	 if (IsValidName(charname)==false or IsValidName(buddyname)==false) then
			-- output other error message
			return false;
	 end

	 local ts;
	 if (timestamp) then
			ts = timestamp;
	 else
			ts = BuddyName:GetTimestamp();
	 end
	 self.db.realm.char[charname]= {
			["buddy"] = buddyname,
			["guild"] = "", -- TODO
			["timestamp"] = ts
	 }
	 return true
end

-- returns true if updated, false if not updated
function BuddyName:UpdateIfNewer(charname, buddyname, timestamp)
	 if (charname~=nil and charname~="" and buddyname~=nil and timestamp~=nil) then
			if (not self.db.realm.char[charname]) then -- entry does not exist at all
				 self:Print(string.format("Adding new char from synchronization: %s => %s", charname, buddyname)); -- debug
				 BuddyName:addToDB(charname,buddyname,timestamp); -- TODO: return value ignored ATM
				 return true;
			elseif (self.db.realm.char[charname].timestamp < timestamp) then -- Check if received timestamp is newer
				 self:Print(string.format("Updating char from synchronization: %s => %s", charname, buddyname)); -- debug
				 BuddyName:addToDB(charname,buddyname,timestamp); -- TODO: return value ignored ATM
				 return true;
			end
	 end
	 return false;
end

--[[ Deletes an entry from the DB if it exists. Depending on the deleteCompletely flag,
--   an existing entry may either be deleted completely (true) or just set to "" (false).
--   The latter leaves the actual entry intact, resulting in its spreading over the guild
--   synchronization mechanism. Therefore, it is possible to delete this entry from other
--   guild members DBs as well.
--   Returns:
--     1. Boolean: true if deletion was successful, false else
--     2. String: buddy name if deletion was successful, nil else (e.g. char was not in DB)
--]]
function BuddyName:deleteFromDB(charname, deleteCompletely)
	 -- is this char in the DB?
	 if (not BuddyName.db.realm.char[charname]) then
			return false,nil
	 end

	 -- perform deletion and set return value
	 local buddyname = BuddyName.db.realm.char[charname].buddy
	 if (deleteCompletely==true) then
			BuddyName.db.realm.char[charname]=nil
	 else
			BuddyName.db.realm.char[charname].buddy=""
			BuddyName.db.realm.char[charname].timestamp=BuddyName:GetTimestamp();
	 end

	 return true,buddyname
end

function BuddyName:FixDB(version)
	 -- TODO: also check for entries missing required items. For each DBv

	 if (version==2) then
			-- remove all char names and buddy names that are not valid. May have happend during the RealID test phase
			for p,_ in pairs(self.db.realm.char) do

				 local inDB, buddy = BuddyName:IsInDB(p);
				 if (IsValidName(p)==false or (inDB==true and IsValidName(buddy)==false)) then
						self.db.realm.char[p] = nil;
						self:Print(string.format("Removed broken entry from DB"));
				 end
			end
	 end
end

--[[ Transforms old DB entries to newer versions
--   Returns: nil
--]]
function BuddyName:UpdateDB()
	 if (self.db.realm.version==1) then
			self:Print(string.format(L["UPDATING_DB"], 2));

			-- perform updates to v2
			for p,v in pairs(self.db.realm.char) do
				 if (self.db.realm.char[p]) then
						self.db.realm.char[p] = {
							 buddy = v,
							 guild = "", -- TODO
							 timestamp = BuddyName:GetTimestamp(),
						}
				 end
			end

			self.db.realm.version=2
			self:Print(L["UPDATE_OK"])
	 end

	 --[[
	 if (self.db.realm.version==2) then
			self:Print(string.format(L["UPDATING_DB"], 3));
			-- perform update to v3 here
			self.db.realm.version=3
			self:Print(L["UPDATE_OK"])
	 end
	 ]]
end

--[[ -- -------------- -- ]]
--[[ -- Message output -- ]]
--[[ -- -------------- -- ]]

--[[ Hooked-in version of the AddMessage function.
--   Will check if message contains matching character link,
--   and insert buddy name if match was found.
--   Returns: nil
--]]
function BuddyName:AddMessage(frame, msg, ...)
	 local pstart,pend,sender,type; -- variables used in string matching below

	 -- match chat channels, including guild, whisper, etc
	 pstart,pend,sender,type = string.find(msg, "|Hplayer:(%a+):%d+:(%u+)[^|]*|h%[[^%]]+%]"); -- TODO: try to match against |h..|h instead of [..]
	 if (pstart) then -- something matching was found, else nil
			local replace,buddy = BuddyName:IsInDB(sender) -- char in DB?
			if (replace==true) then
				 if (chatChannels[type]==true or -- msg occured in one of the monitored channels
						string.find(msg,L["ANNOUNCE_ONLINE"]) or -- player has come online
						string.find(msg,L["ANNOUNCE_ACHIEVEMENT"]) ) -- player earned an achievement
				 then
						msg=string.gsub(msg, "(%[[^%]]*)("..sender..")([^%[]*%])", '%1%2:'..buddy..'%3',1); -- only replace first occurrence
				 end
			end
	 else
			-- match people coming online or earning achievements
			-- test: always replace first occurrence of player here
			pstart,pend,sender=string.find(msg,"|Hplayer:(%a+)|h%[%a+%]|h"); -- Compatible to Prat here?
			if (pstart) then
				 local replace,buddy = BuddyName:IsInDB(sender);
				 if (replace==true) then
						msg=string.gsub(msg,"(|Hplayer:%a+|h%[[^%]]*)("..sender..")([^%]]*]|h)", "%1%2:"..buddy.."%3" ,1);
				 end
			else
				 -- match people going offline
				 pstart,pend,sender=string.find(msg, L["ANNOUNCE_OFFLINE"]);
				 if (pstart) then
						local replace,buddy = BuddyName:IsInDB(sender) -- char in DB?
						if (replace) then
							 msg=string.gsub(msg, sender, sender..":"..buddy, 1);
						end
				 end
			end
	 end
	 
	 -- call original AddMessage() function
	 self.hooks[frame]["AddMessage"](frame, msg, ...);
end

--[[ Event Handler that is currently not in use. Perhaps in future version agains.
--   Returns: nil
--]]
local function eventHandler(self, event, ...)
	 if (BN_Active == false) then
			return;
	 end

	 if (registerEvents[event]==true) then
			--print("Registered event of type "..event); -- debug
	 end
end

--[[ -- ----------------- -- ]]
--[[ -- AddOn interaction -- ]]
--[[ -- ----------------- -- ]]

--[[ Process slash commands (unexpected, eh?)
--   Returns: nil
--]]
function BuddyName:SlashProcessor(input)
	 -- 'input' contains whatever follows the slash command

	 -- TODO: respect text contained in "..." or '...', split correctly

	 -- split input string, ignore additional whitespace
	 local tbl = {}
	 for v in string.gmatch(input, "[^ ]+") do
			tinsert(tbl, v)
	 end

	 -- if no additional parameters => print out help text
	 if (#tbl == 0 or (tbl[1]~=nil and tbl[1]=="help")) then
			self:Print(L["HELP"])
			return;
	 end

	 local command = tremove(tbl,1);

	 -- Activate additional functionality
	 if (command=="on") then
			self:Enable()
			self:Print(L["ACTIVATED"]);
			return
	 end

	 -- Disable additional functionality
	 if (command=="off") then
			self:Disable()
			self:Print(L["DEACTIVATED"]);
			return
	 end

	 if (not BN_Active) then
			self:Print(L["NOT_ACTIVE"])
			return
	 end

	 -- add another person to the list
	 if (command=="add") then
			if (#tbl~=2) then
				 self:Print("Add: "..L["WRONG_NUMBER_OF_ARGUMENTS"])
			end
			local successful = BuddyName:addToDB(tbl[1], tbl[2]);
			if (successful==true) then
				 self:Print(string.format(L["BUDDY_ADD_OK"], tbl[1], tbl[2]));
			else
				 self:Print(string.format("Could not add %s => %s to DB", tbl[1], tbl[2]));
			end
			return
	 end

	 -- add another person to the list
	 if (command=="del" or command=="rm") then
			if (#tbl~=1) then
				 self:Print("Del: "..L["WRONG_NUMBER_OF_ARGUMENTS"])
			end
			local successful,buddyname = BuddyName:deleteFromDB(tbl[1], false)
			if (not successful) then
				 self:Print(string.format(L["BUDDY_DEL_ERROR"], tbl[1]))
				 return
			end
			self:Print(string.format(L["BUDDY_DEL_OK"], tbl[1], buddyname));
			return
	 end

	 -- list chars from database
	 if (command=="list") then
			self:Print(string.format(L["BUDDY_LIST_HEAD"],currentRealm));

			for p,v in pairs(BuddyName.db.realm.char) do
				 local char = p;
				 local buddy = BuddyName.db.realm.char[p].buddy -- DB v2
				 if (buddy and buddy~="") then
						self:Print(string.format(L["BUDDY_LIST_ENTRY"], char, buddy));
				 end
			end

			-- TODO:
			-- Distinguish between all, guild, non-guild
			-- => Additional command arguments
			return
	 end

	 -- to test the communication properties
	 if (command=="comm") then
			if (commActive==true) then
				 commActive=false;
				 self:Print("Switched off communication");
			else
				 commActive=true;
				 self:Print("Switched on communication");
			end
			return;
	 end

	 -- to test the communication properties
	 if (command=="sync") then
			if  (commActive==true) then
				 self:Print("Initiating guild broadcast");
				 BuddyName:BroadcastAllToGuild();
			else
				 self:Print("Guild broadcast not initiated as communication is not active");
			end
			return;
	 end

	 -- toggle the output of debug statements
	 if (command=="debug") then
			if (debugMode==true) then
				 debugMode=false;
				 self:Print("Switched off debug messages");
			else
				 debugMode=true;
				 self:Print("Switched on debug messages");
			end
			return;
	 end

	 self:Print(string.format(L["UNKNOWN_COMMAND"], command));
end

--[[ -- ------------------------- -- ]]
--[[ -- AddOn chat communication  -- ]]
--[[ -- ------------------------- -- ]]

--[[ Extracts the header from a given frame string without modifying the frame
--   itself. Does not check whether the header contains valid data, just
--   performs format checking.
--   Returns:
--     1. table { protocol, compression, length } if successfull, length refers
--        to header length;
--        nil if not successful.
--]]
local function CommGetFrameHeader(frame)
	 local hstart,hlength,hprotocol,hcompression = string.find(frame, COMM.PATTERNS.SEARCH.HEADER);

	 --tinsert(BuddyName.db.realm["Comm"].msg, frame); -- debug

	 local header = nil;
	 if (hstart) then
			header = {
				 protocol = tonumber(hprotocol), -- yes, tonumber() is necessary here
				 compression = tonumber(hcompression),
				 length = hlength,
			};
			debug("Header: "..allToString(header.protocol, header.compression, header.length));
	 end

	 return header;
end

--[[ Checks if the header uses the current communication protocol,
--   and if the other values included in the header make sense.
--   Method will output error messages unless told via argument
--   to stay silent. Default is silent=false.
--   Returns:
--     1. true if header is valid, false else
--]]
local function CommHeaderIsValid(header, silent)
	 if (not silent) then
			silent = false;
	 end

	 -- was there even an argument provided?
	 if (header==nil) then
			if (silent==false) then
				 debug("No header provided");
			end
			return false;
	 end

	 -- all required fields present and of right type?
	 if (header.protocol==nil or tonumber(header.protocol)==nil or
	     header.compression== nil or tonumber(header.compression)==nil or
	     header.length==nil or header.length<1) then
			if (silent==false) then
				 debug("Header has missing fields or non-numerical values");
			end
			return false;
	 end

	 -- protocol too old?
	 if (header.protocol < COMM.PROTOCOL) then
			if (silent==false) then
				 debug("Header uses old communication protocol");
				 -- silently ignore old clients for now
				 -- perhaps will still be processed later
				 -- e.g. by whispering to sender with a hint to update
			end
			return false;
	 end

	 -- protocol too new?
	 if (header.protocol > COMM.PROTOCOL) then
			if (silent==false) then
				 if (warnedAboutUpdate==false) then
						-- TODO: change from debug to real output
						debug("A synchronization message was received from a newer client, please update");
						warnedAboutUpdate=true;
				 end
			end
			return false;
	 end

	 -- do the values in these fields make sense?
	 -- compare against compression type with lowest and highest number
	 if (header.compression<COMM.HEADER.COMPRESSION.NONE or
	     header.compression>COMM.HEADER.COMPRESSION.LIBCOMPRESS) then
			if (silent==false) then
				 -- TODO: change from debug to real output
				 debug("Header uses an unknown compression type: "..header.compression);
			end
			return false
	 end

	 return true;
end

--[[ Takes an uncompressed data section as string, without delimiting "|".
--   Then determines data section type and splits up content according to
--   the section types format definition. The frame and its content is
--   returned in form of a table. If the given string is malformed, the
--   resulting table contains COMM.SECTION_TYPE.UNKNOWN as section type.
--   Returns:
--     1. Table = {
--        	 type = sectionType according to COMM.SECTION_TYPE.*
--        	 entry = {}: indexed table of entries in section in order
--        }
--]]
local function CommDeserializeSection(sectionString)
	 if (not sectionString) then
			debug("nil string given to CommDeserializeSection");
			return { type = COMM.SECTION_TYPE.UNKNOWN, entry={} };
	 end

	 debug("sectionString: \""..sectionString.."\"");

	 -- check if provided string matches common data section pattern
	 local pstart,pend,secType = string.find(sectionString, COMM.PATTERNS.SEARCH.SECTION_TYPE);
	 if (not pstart) then
			debug("Found unknown section type");
			return { type = COMM.SECTION_TYPE.UNKNOWN, entry={} };
	 else
			secType=tonumber(secType); -- convert from string to numerical
	 end

	 -- base structure for resulting table
	 local section = {
			type  = secType,
			entry = {},
	 };

	 -- split up remaining string after section type indicator
	 local entries={ strsplit(COMM.SEP.ENTRY, strsub(sectionString,pend+1)) };

	 -- section of type ENTRY_SET
	 if (secType==COMM.SECTION_TYPE.ENTRY_SET) then
			debug("Processing ENTRY_SET");
			for i=1, #entries do -- process all included chars
				 debug("Processing entry "..i);
				 if (entries[i]~=nil and entries[i]~="") then
						local entry = { strsplit(COMM.SEP.FIELD, entries[i]); };
						if (#entry ~= 3) then -- correct number of fields in entry?
							 print("Malformed ENTRY_SET message");
						else
							 entry[3] = tonumber(entry[3]);
							 if (entry[3]==nil) then -- nil if conversion was not possible
									print("Malformed ENTRY_SET message");
							 else
									debug("Inserted entry in section table");
									tinsert(section.entry, entry);
							 end
						end
				 end
			end

	 -- section of type HASH_PROPAGATION
	 elseif (secType==COMM.SECTION_TYPE.HASH_PROPAGATION) then
			debug("Processing HASH_PROPAGATION");
			if (#entries ~= 4) then
				 print("Malformed HASH_PROPAGATION message");
			else
				 section.entry = {
						dataDomain = tonumber(entries[1]),
						hash       =          entries[2],
						entries    = tonumber(entries[3]),
						timestamp  = tonumber(entries[4]),
				 };

				 -- TODO: include validity checks
			end

	 -- section of type DATA_REQUEST
	 elseif (secType==COMM.SECTION_TYPE.DATA_REQUEST) then
			debug("Processing DATA_REQUEST");
			if (#entries ~= 4) then
				 print("Malformed DATA_REQUEST message");
			else
				 section.entry = {
						dataDomain = tonumber(entries[1]),
						partial    = tonumber(entries[2]),
						response   = tonumber(entries[3]),
						receiver   =          entries[4],
				 };

				 -- TODO: include validity checks
			end
	 else
			debug("Unknown section type");
			return { type = COMM.SECTION_TYPE.UNKNOWN };
	 end

	 return section;
end

--[[ Takes a complete frame as string, checks the header, and decodes the frames'
--   other data sections according to their format. The result is returned as a
--   table consisting of the header and an ordered list of data sections. In
--   case of a malformed frame header, the header table in the result will be nil
--   and the data sections will not be processed. In case of malformed data
--   sections, they will not be included in the section table, but a warning
--   is printed instead.
--   In case of a header with invalid data or if the protocol is newer,
--   table with header=nil also will be returned as well.
--   Returns:
--     1. Table = {
--        	 header = table according to CommGetFrameHeader()
--        	 section = {}: indexed table of data sections in order;
--        	               sections are formated as tables according to
--        	               function CommDeserializeSection()
--        }
--]]
local function CommDecodeFrame(frameString)
	 local frame = {
			header = CommGetFrameHeader(frameString),
			section = {},
	 };

	 -- check header validity.
	 debug(frame.header); -- remove me!
	 if (CommHeaderIsValid(frame.header,false)==false) then
			return { header = nil, section = nil };
	 end

	 -- remove header from frame string, including section delimiter "|"
	 frameString = string.sub(frameString, frame.header.length+1);

	 -- decompress remaining data sections if necessary
	 if (frame.header.compression==COMM.HEADER.COMPRESSION.NONE) then
			debug("No compression on received frame");
			-- no compression, do nothing
	 elseif (frame.header.compression==COMM.HEADER.COMPRESSION.LIBCOMPRESS) then
			debug("LibCompress compression on received frame");
			--debug(string.format("Decompress before: %d",strlen(frameString)));

			frameString = Compressor:Decompress(frameString);

			--debug(string.format("Decompress after: %d",strlen(frameString)));
	 else -- no valid compression indicator found, cancel
			print("No known compression type on frame: "..frame.header.compression);
			return frame;
	 end

	 -- split frame string into data sections, deserialize them and put into table
	 local sections = { strsplit(COMM.SEP.SECTION, frameString) };
	 for i=1, #sections do
			debug("Processing section "..i);
			if (sections[i]~=nil and sections[i]~="") then -- TODO: check if nil values can orrur here
				 local section = CommDeserializeSection(sections[i]);
				 if (section.type~=COMM.SECTION_TYPE.UNKNOWN) then
						tinsert(frame.section, section);
				 end
			end
	 end
	 debug("All sections processed");

	 return frame;
end

--[[ Automatically called by AceComm3 if a message with matching prefix
--   was received.
--   Returns: nil
--]]
function BuddyName:OnCommReceived(prefix, frameString, channel, sender)
	 if (commActive==false) then -- for debugging purposes
			return;
	 end

	 debug("Received frame via comm");

	 -- test if frame is from different addon or yourself
	 --if (prefix~=COMM.SYNC_PREFIX or sender==GetUnitName("player",false)) then
	--		return;
	 --end

	 local frame = CommDecodeFrame(frameString);

	 if (frame.header==nil) then
			self:Print("Received frame with malformed header by "..sender);
			return;
	 end

	 debug("Received message:");
	 debug(allToString(frame));

	 -- some debug information
	 local source;
	 if (channel=="GUILD") then
			source = "guild ("..sender..")";
	 elseif (channel=="WHISPER") then
			source = "player "..sender;
	 else
			source = "Unknown";
	 end
	 self:Print(string.format("Received frame from %s", source));

	 -- process all received sections
	 for i=1, #(frame.section) do
			local section = frame.section[i];
			debug("Processing data section of type "..section.type);

			-- process section of type ENTRY_SET
			if (section.type == COMM.SECTION_TYPE.ENTRY_SET) then
				 for j=1, #(section.entry) do
						local char  = section.entry[j][1];
						local buddy = section.entry[j][2];
						local ts    = section.entry[j][3];
						debug(string.format("Processing entry %s:%s:",char,buddy)..ts);
						BuddyName:UpdateIfNewer(char,buddy,ts);
				 end

			-- process section of type HASH_PROPAGATION
			elseif (section.type == COMM.SECTION_TYPE.HASH_PROPAGATION) then
				 -- TODO: fill me

			-- process section of type DATA_REQUEST
			elseif (section.type == COMM.SECTION_TYPE.DATA_REQUEST) then
				 -- TODO: fill me

			end
	 end
end

function BuddyName:BroadcastAllToGuild()
	 if (commActive==false) then -- for debugging purposes
			return;
	 end

	 if (PlayerIsInGuild()) then
			local header=COMM.PATTERNS.BUILD.HEADER;

			local frame = COMM.PATTERNS.BUILD.ENTRY_SET_STUB;
			local entries = 0;

			for p,v in pairs(self.db.realm.char) do
				 if (self.db.realm.char[p] and UnitIsInMyGuild(p)) then
						self:Print(string.format("Sharing guild member %s",p)); -- debug
						local newEntry=strjoin(COMM.SEP.FIELD, p, self.db.realm.char[p].buddy,self.db.realm.char[p].timestamp);
						frame=frame..COMM.SEP.ENTRY..newEntry;
						entries=entries+1;
				 end
			end

			if (entries==0) then
				 print("No entries to broadcast, canceling sync");
				 return;
			end

			-- always compress for now
			--debug(string.format("Compress before: %d",strlen(frame)));
			frame=Compressor:Compress(frame);
			--debug(string.format("Compress after: %d",strlen(frame)));

			-- build final output string
			frame=header..frame;

			BuddyName:SendCommMessage(COMM.SYNC_PREFIX, frame, "GUILD");
	 end
end

--[[ -- --------------------------------------------- -- ]]
--[[ -- Basic functions automatically called by Ace3  -- ]]
--[[ -- --------------------------------------------- -- ]]

--[[ Executed only once on startup
--   Returns: nil
--]]
function BuddyName:OnInitialize()
	 debug("OnInitialize called");

	 currentRealm = GetRealmName();

	 -- Register Database
	 self.db = LibStub("AceDB-3.0"):New("BuddyName_DB")

	 -- Register Slash commands
	 for p,v in pairs(slashCommands) do
			if (v == true) then
				 self:RegisterChatCommand(p, "SlashProcessor")
			end
	 end

	 -- Create frame for events to hook up to and register battle net events
	 BN_EventFrame = CreateFrame("Frame", "BuddyNameFrame");
	 BN_EventFrame:SetScript("OnEvent", eventHandler);
	 for p,v in pairs(registerEvents) do
			if (v == true) then
				 BN_EventFrame:RegisterEvent(p);
			end
	 end

	 -- Register for receiving communication for synchronization purposes
	 BuddyName:RegisterComm(COMM.SYNC_PREFIX);
end


--[[ Executed each time the addon is switched on
--   Returns: nil
--]]
function BuddyName:OnEnable()
	 debug("OnEnable called");

	 if (BN_Active == true) then
			return -- we're already active
	 else
			BN_Active = true;
	 end

	 -- check if DB already exists, or if it is old and has to be converted
	 -- not done earlier because has to be called after PLAYER_ENTERING_WORLD (because of UpdateDB())
	 if (self.db.realm.version==nil) then -- if no version number, then either fresh db or very old. In each case, reset realm DB.
			self.db.realm.version=BN_DB_Version -- set current version number
			self.db.realm.char = nil; -- set empty char table
			self.db.realm.char = {};
	 elseif (self.db.realm.version < BN_DB_Version) then -- update old database
			BuddyName:UpdateDB();
	 elseif (self.db.realm.version > BN_DB_Version) then -- database is newer than BuddyName
			self:Print(L["UPDATE_BUDDYNAME"]);
			BN_Active = false;
			return;
	 end

	 -- Apply fixes for broken DB entries
	 if (self.db.realm.version==2) then
			BuddyName:FixDB(2);
	 end

	 -- Hook into AddMessage
	 for i = 1, NUM_CHAT_WINDOWS do
			if (i~=2) then -- exclude combat log, better check against type than number
				 self:RawHook(_G["ChatFrame"..i], "AddMessage", "AddMessage", true);
			end
	 end

	 -- Add one-shot timer to check for battle.net friends in a few seconds
	 --BN_Timer["BattleNet"] = self:ScheduleTimer("AddBattleNetFriends", 5);
end


--[[ Executed each time the addon is switched off
--   Returns: nil
--]]
function BuddyName:OnDisable()
	 debug("OnDisable called");

	 if (BN_Active == false) then
			return -- we're already inactive
	 else
			BN_Active = false;
	 end

	 -- Unhook from AddMessage
	 self:Unhook("AddMessage");

	 -- Cancel battle.net timer if it has not already fired
	 --self:CancelTimer(BN_Timer["BattleNet"], true); -- true: don't raise exception if timer is invalid
end
